! ===========================================================================
module WMAP_eebb_lowl

! This code calculates the likelihood function of Q/U 
! polarization for EE signal plus noise. 
!
! Use combined 3yr data.
! - Combined Ninv and weighted maps have been generated by
! maps/weight_degrade_coadd_p2.f90
!
! - NinvPlNinv has been generated by 
! ninv/generate_ninvplninv_r3_p2.f90
!
! E. Komatsu, October 4, 2005
! Incorporated into likelihood code October 7th RB.
! BB has been added, November 6th EK.
!
! - Implementation of David's likelihood
! Ek, December 16, 2005
! ===========================================================================

  Implicit None
  integer :: nsmax,nlmax,np,mp
  REAL, allocatable, dimension(:,:,:,:), Target :: ninvplninv
  INTEGER, allocatable, dimension(:) :: ngood
  REAL, allocatable, dimension(:,:) :: Dp0
  REAL, allocatable, dimension(:) :: m_r3,w_r3,p_r3
  COMPLEX, allocatable, dimension(:,:), Target :: alm_tt,NinvY
  REAL, allocatable, dimension(:) :: wl
contains

!===========================================
subroutine eebb_lowl_like_setup
!===========================================

  use wmap_util
  use wmap_options, only : WMAP_data_dir

  implicit none

  character(LEN=2) :: rlz
  integer,parameter :: ires = 3
  REAL, dimension(:),allocatable :: T,N,Mask_R3
  CHARACTER(len=3) :: da(10),sband
  INTEGER :: ReadStatus
  CHARACTER(len=128) :: qfile,ufile,maskfile,filename(0:6),eebbdir
  DOUBLE PRECISION, allocatable, dimension(:,:) :: window_th

  INTEGER :: dum,nsmax,nlmax,l,ip,jp,m,lun,stat
  REAL, allocatable, dimension(:,:), Target :: NinvQUr3
  logical ::yes

  Complex (Kind=4), Dimension(:,:),   Pointer :: Cptr2
  Real (Kind=4),    Dimension(:,:),   Pointer :: Rptr2
  Real (Kind=4),    Dimension(:,:,:), Pointer :: Rptr3

  INCLUDE 'read_archive_map.fh'
  Include 'read_fits.fh'

!------------------------------
! Set initial parameters
!-----------------------------

  da=(/'K1','Ka','Q1','Q2','V1','V2','W1','W2','W3','W4'/)
  da(2)='Ka1'


  nsmax = 2**ires
  nlmax = 3*nsmax-1 ! use Nyquist sampling rather than 2*nsmax
  np = 12*nsmax**2

  eebbdir = WMAP_data_dir

  filename(0)=trim(eebbdir)//'ninv/masked_ee_ninvplninv_qu_r3_p2_norm_corrected_3yr.QV.fits'
  filename(1)=trim(eebbdir)//'ninv/masked_bb_ninvplninv_qu_r3_p2_norm_corrected_3yr.QV.fits'
  filename(2)=trim(eebbdir)//'ninv/masked_ninv_qu_r3_p2_norm_corrected_3yr.QV.fits'
  filename(3)=trim(eebbdir)//'maps/wt_r3_p2_3yr.QV.map_q'
  filename(4)=trim(eebbdir)//'maps/wt_r3_p2_3yr.QV.map_u'
  filename(5)=trim(eebbdir)//'alm/alm_tt_fs_r9_ilc_nopixwin_3yr.dat'
  filename(6)=trim(eebbdir)//'ninv/masked_ninvy_qu_r3_p2_norm_corrected_3yr.QV.fits'
 
!------------------------------
! read in res3 mask
!------------------------------

  allocate(mask_r3(0:np-1))
  allocate(T(0:np-1))
  allocate(N(0:np-1))

  maskfile=trim(eebbdir)//'mask/mask_r3_p06.fits'
  inquire(file= maskfile,exist = yes)
  if(.not.yes)then
     write(*,*)"eebb maskfile not found",trim(maskfile)
  endif
  CALL READ_ARCHIVE_MAP(maskfile,T,Mask_R3,dum,ReadStatus)
  if (ReadStatus/=0) then
     print*,'unable to read in mask',trim(maskfile)
     stop
  endif

  ALLOCATE(ngood(0:np-1))
  mp = 0
  do ip=0,np-1
     if (Mask_R3(ip)/=0) then
        ngood(mp) = ip

        mp = mp + 1

     endif
  enddo

  !------------------------------
  ! read in N^{-1}P_l N^{-1} at res3
  !------------------------------

  ALLOCATE(ninvplninv(0:nlmax,0:2*np-1,0:2*np-1,2), stat=stat)
  If (stat .NE. 0) Then
    Print *, 'Memory allocation error for ninvplninv'
    Stop
  End If

  Rptr3 => ninvplninv(:,:,:,1)
  Call Read_FITS_Real_3D (filename(0), Rptr3, stat)
  If (stat .NE. 0) Then
    Print *, 'Error ', stat, ' while reading ', trim(filename(0))
    Stop
  End If

  Rptr3 => ninvplninv(:,:,:,2)
  Call Read_FITS_Real_3D (filename(1), Rptr3, stat)
  If (stat .NE. 0) Then
    Print *, 'Error ', stat, ' while reading ', trim(filename(1))
    Stop
  End If

  !------------------------------
  ! read in N^{-1} at res3
  !------------------------------

  ALLOCATE(NinvQUr3(0:1535,0:1535), stat=stat)
  If (stat .NE. 0) Then
    Print *, 'Memory allocation error for NinvQUr3'
    Stop
  End If

  Rptr2 => NinvQUr3
  Call Read_FITS_Real_2D (filename(2), Rptr2, stat)
  If (stat .NE. 0) Then
    Print *, 'Error ', stat, ' while reading ', trim(filename(2))
    Stop
  End If

  !------------------------------
  ! read in maps at res3
  !------------------------------

  ALLOCATE(w_r3(0:2*mp-1),m_r3(0:2*mp-1),p_r3(0:2*mp-1), stat=stat)
  If (stat .NE. 0) Then
    Print *, 'Memory allocation error for w_r3, m_r3, or p_r3'
    Stop
  End If
  qfile=filename(3)
  CALL READ_ARCHIVE_MAP(qfile,T,N,np,ReadStatus)
  if(ReadStatus/=0)then
     print*,'unable to read in '//trim(qfile)
     stop
  endif

  do ip=0,mp-1
     w_r3(ip) = T(ngood(ip))*Mask_R3(ngood(ip))
  enddo

  ufile=filename(4)
  CALL READ_ARCHIVE_MAP(ufile,T,N,np,ReadStatus)
  if(ReadStatus/=0)then
     print*,'unable to read in '//trim(ufile)
     stop
  endif

  do ip=0,mp-1
     w_r3(mp+ip) = T(ngood(ip))*Mask_R3(ngood(ip))
  enddo

  ALLOCATE(Dp0(0:2*mp-1,0:2*mp-1), stat=stat)
  If (stat .NE. 0) Then
    Print *, 'Memory allocation error for Dp0'
    Stop
  End If
  do ip=0,mp-1
     do jp=0,mp-1
        Dp0(ip,jp) = NinvQUr3(ngood(ip),ngood(jp))
        Dp0(ip,mp+jp) = NinvQUr3(ngood(ip),np+ngood(jp))
        Dp0(mp+ip,jp) = NinvQUr3(np+ngood(ip),ngood(jp))
        Dp0(mp+ip,mp+jp) = NinvQUr3(np+ngood(ip),np+ngood(jp))
     enddo
  enddo

  !------------------------------
  ! read in alm_tt 
  !------------------------------
  ALLOCATE(alm_tt(0:nlmax,0:nlmax), stat=stat)
  If (stat .NE. 0) Then
    Print *, 'Memory allocation error for alm_tt'
    Stop
  End If
  call get_free_lun( lun )
  open(lun,file=filename(5),status='old')
  do l=0,nlmax
     do m=0,l
        read(lun,*)alm_tt(l,m)
     enddo
  enddo
  close(lun)

  !------------------------------
  ! read in Ninv Y
  !------------------------------
  ALLOCATE(NinvY(0:1535,300), stat=stat)
  If (stat .NE. 0) Then
    Print *, 'Memory allocation error for NinvY'
    Stop
  End If

  Cptr2 => NinvY
  Call Read_FITS_Complex_2D_LM (filename(6), Cptr2, stat, IndFmt=2)
  If (stat .NE. 0) Then
    Print *, 'Error ', stat, ' while reading ', trim(filename(6))
    Stop
  End If

!
! read in pixel window
!
  ALLOCATE(wl(0:nlmax),window_th(0:4*nsmax,1))
  CALL Get_Pixel_Window(nsmax,window_th)
  wl(0:nlmax) = window_th(0:nlmax,1)
  DEALLOCATE(window_th)

  deallocate(T,N,Mask_R3)
  DEALLOCATE(NinvQUr3)
  return

end subroutine eebb_lowl_like_setup

!===========================================
subroutine eebb_lowl_likelihood(nlmaxin,clttin,cltein,cleein,clbbin,chisq_r3,lndet)
!=======================================

  use wmap_util

  IMPLICIT NONE

  REAL(8), dimension(2:*),intent(in) :: clttin,cltein,cleein,clbbin
  real(8),intent(out) ::lndet,chisq_r3 ! like
  integer, intent(in) ::nlmaxin
  REAL, allocatable, dimension(:,:) :: NinvSNinv
  real,allocatable,dimension(:,:) ::CQU
  REAL, allocatable, dimension(:,:) :: Dp 
  real,allocatable,dimension(:) ::clee,clbb,cltt,clte
  REAL :: SDOT !chisq_r3,SDOT,lndet
  INTEGER :: ip,jp,nlmax,info,l,m,i
  integer :: t_start,t_end,crate,cmax

  nlmax = 23
  if(nlmaxin.ne.nlmax)then
     write(*,*)"need nlmax",nlmax,", in eebb likelihood, currently",nlmaxin
     stop
  endif

  allocate( Cltt(0:nlmax),Clte(0:nlmax),Clee(0:nlmax),Clbb(0:nlmax) )
  Cltt(0:1)=0.d0
  Clte(0:1)=0.d0
  Clee(0:1)=0.d0
  Clbb(0:1)=0.d0
  do l=2,nlmax
     Cltt(l) = Clttin(l)/dble(l*(l+1))*(2.*3.14159)*1.d-6*wl(l)**2.
     Clte(l) = Cltein(l)/dble(l*(l+1))*(2.*3.14159)*1.d-6*wl(l)**2.
     Clee(l) = Cleein(l)/dble(l*(l+1))*(2.*3.14159)*1.d-6*wl(l)**2.
     Clbb(l) = Clbbin(l)/dble(l*(l+1))*(2.*3.14159)*1.d-6*wl(l)**2.
  enddo

  !------------------------------
  ! compute [N^{-1}S N^{-1} + N^{-1}]^{-1}
  !------------------------------
  ALLOCATE(CQU(0:2*mp-1,0:2*mp-1))
  ALLOCATE(NinvSNinv(0:2*mp-1,0:2*mp-1))
  ALLOCATE(Dp(0:2*mp-1,0:2*mp-1))

  lndet = 0.
  NinvSNinv = 0.

!
! fill only the upper triangular part of N^{-1}SN^{-1}.
!

  do ip=0,mp-1
     do jp=ip,mp-1
        NinvSNinv(ip,jp) = &
             +SDOT(nlmax-1,clee(2:)-clte(2:)**2./cltt(2:),1,ninvplninv(2:,ngood(ip),ngood(jp),1),1) &
             +SDOT(nlmax-1,clbb(2:),1,ninvplninv(2:,ngood(ip),ngood(jp),2),1)
        
        NinvSNinv(ip,mp+jp) = &
             +SDOT(nlmax-1,clee(2:)-clte(2:)**2./cltt(2:),1,ninvplninv(2:,ngood(ip),np+ngood(jp),1),1)&
             +SDOT(nlmax-1,clbb(2:),1,ninvplninv(2:,ngood(ip),np+ngood(jp),2),1) 

        NinvSNinv(mp+ip,mp+jp) = &
             +SDOT(nlmax-1,clee(2:)-clte(2:)**2./cltt(2:),1,ninvplninv(2:,np+ngood(ip),np+ngood(jp),1),1) &
             +SDOT(nlmax-1,clbb(2:),1,ninvplninv(2:,np+ngood(ip),np+ngood(jp),2),1)
     enddo
     do jp=0,ip-1
        NinvSNinv(ip,mp+jp) = NinvSNinv(ip,mp+jp) &
             +SDOT(nlmax-1,clee(2:)-clte(2:)**2./cltt(2:),1,ninvplninv(2:,ngood(ip),np+ngood(jp),1),1)&
             +SDOT(nlmax-1,clbb(2:),1,ninvplninv(2:,ngood(ip),np+ngood(jp),2),1) 
     enddo
  enddo

  Dp = Dp0 + NinvSNinv 

  DEALLOCATE(NinvSNinv)

  CALL SPOTRF('U',2*mp,Dp,2*mp,info) 
  IF(info.NE.0)then
	call wmap_likelihood_error( 'eebb: bad spotrf', info )
	chisq_r3 = 0d0
	lndet = 0d0
	return

  endif

  do ip=0,2*mp-1
     lndet = lndet + 2.*log(real(Dp(ip,ip)))
  enddo

   CALL SPOTRI('U',2*mp,Dp,2*mp,info)
  IF(info.NE.0)then
	call wmap_likelihood_error( 'eebb: bad spotri', info )
	chisq_r3 = 0d0
	lndet = 0d0
	return

  endif

  CQU(:,:) = Dp 

  DEALLOCATE(Dp)


  !-----------------------------------
  ! calculate the predicted QU at res3
  !-----------------------------------
  p_r3 = 0.
  do ip=0,mp-1
     i = 3
     do l=2,23
        i = i+1
        p_r3(ip)    = p_r3(ip)    &
		+ clte(l)/cltt(l)*wl(l)*alm_tt(l,0)*NinvY(ngood(ip),i)
        p_r3(ip+mp) = p_r3(ip+mp) &
		+ clte(l)/cltt(l)*wl(l)*alm_tt(l,0)*NinvY(ngood(ip)+768,i)
        do m=1,l
           i = i+1
           p_r3(ip) = p_r3(ip)&
		+ clte(l)/cltt(l)*wl(l)*( alm_tt(l,m)*NinvY(ngood(ip),i)&
		                  +conjg(alm_tt(l,m)*NinvY(ngood(ip),i)) )
           p_r3(ip+mp) = p_r3(ip+mp)&
                + clte(l)/cltt(l)*wl(l)*( alm_tt(l,m)*NinvY(ngood(ip)+768,i)&
		                  +conjg(alm_tt(l,m)*NinvY(ngood(ip)+768,i)) )
        enddo
     enddo 
  enddo

  CALL SSYMV('U',2*mp,1.,CQU(:,:),2*mp,w_r3-p_r3,1,0.,m_r3,1)
  chisq_r3 = SDOT(2*mp,m_r3,1,w_r3-p_r3,1) 


  DEALLOCATE(CQU)

  chisq_r3 = chisq_r3/2.
  lndet =lndet/2

  return

END subroutine EEBB_LOWL_LIKELIHOOD

!================================================================-
SUBROUTINE GET_PIXEL_WINDOW(nsmax,window_th)
  USE fitstools, ONLY : read_dbintab
  use healpix_types
  use wmap_options, only : wmap_data_dir
  IMPLICIT NONE
  INTEGER, INTENT(IN) :: nsmax
  DOUBLE PRECISION, INTENT(OUT) :: window_th(0:4*nsmax,1:1)
!
  CHARACTER(LEN=80), DIMENSION(1:60)  :: header_PS
  CHARACTER(LEN=128)                  :: infile
  CHARACTER(LEN=128)                  :: windowfile
  CHARACTER(LEN=128)                  :: windowname
  CHARACTER(LEN=128), DIMENSION(4)    :: winfiledir
  CHARACTER(LEN=4)                  :: sstr
  DOUBLE PRECISION :: null_val
  LOGICAL(LGT) :: bad,ok
  INTEGER :: i
!
  write (sstr,'(I4)') nsmax
  sstr=adjustl(sstr)
  windowname='pixel_window_n'&
       & //repeat('0',4-len_trim(sstr))//trim(sstr)//'.fits'

  winfiledir(1)=trim(WMAP_data_dir)//'healpix_data/'
  do i=1,4
     windowfile=trim(winfiledir(i))//trim(windowname)
     inquire(file=windowfile, exist=ok)
     if (ok) exit
  enddo
  if (.not.ok) stop 'Could not find window file.'

  call read_dbintab(windowfile,window_th,4*nsmax+1,1,null_val,bad)

  RETURN
END SUBROUTINE GET_PIXEL_WINDOW

end module WMAP_eebb_lowl
