
module wmap_util

	implicit none

	logical :: wmap_likelihood_ok

	public :: get_free_lun
	public :: wmap_likelihood_ok
	public :: wmap_likelihood_error_init
	public :: wmap_likelihood_error
	public :: wmap_likelihood_error_report

	private

	integer, parameter :: n_errs_max = 10
	character(len=78) :: err_msg(n_errs_max)
	integer :: n_errs, err_code(n_errs_max)

contains

  subroutine get_free_lun( lun )

	implicit none

	integer, intent(out) :: lun

	integer, save :: last_lun = 19
	logical :: used

	lun = last_lun
	do
		inquire( unit=lun, opened=used )
		if ( .not. used ) exit
		lun = lun + 1
	end do

	last_lun = lun

  end subroutine

  subroutine wmap_likelihood_error_init( )

	wmap_likelihood_ok = .true.
	n_errs = 0

  end subroutine

  subroutine wmap_likelihood_error( msg, code )

	character(len=*), intent(in) :: msg
	integer, intent(in) :: code

	wmap_likelihood_ok = .false.

	if ( n_errs < n_errs_max ) then
		n_errs = n_errs + 1
		err_msg(n_errs) = msg
		err_code(n_errs) = code
	else
		print *, '*** error log full'
		stop
	end if

  end subroutine

  subroutine wmap_likelihood_error_report( )

	integer :: i

	print *, '------------------------------------------------------------'
	print *, 'WMAP likelihood evaluation report:'

	if ( wmap_likelihood_ok ) then
		print *, 'no errors'
	else
		print *, 'number of errors = ', n_errs
		do i = 1,n_errs
			print *, ''
			print *, 'error #', i, '::'
			print *, err_msg(i)
			print *, 'error code = ', err_code(i)
		end do
	end if

	print *, '------------------------------------------------------------'

  end subroutine

end module

